#
#  days.rb  --  Time intervals in days
#

require "date"
require "time"

class Date

  def workday_after n
    r = self + n
    case r.wday
      when 0 then r += 1
      when 6 then r += 2
    end
    r
  end

  def quarter
    (month-1) / 3 + 1
  end

  def wstart
    self - wday
  end

  def cwstart
    self - (cwday-1)
  end

end


class Days

  class <<self

    def [] v, b, e = nil
      new v, b, e
    end

  end

  attr_reader :from, :to, :exclusive
  alias exclusive? exclusive
  alias exclude_end? exclusive

  def initialize from, to, exclusive = nil
    @from = mkdate from
    @to   = mkdate to
    @exclusive = exclusive ? true : false
  end

  def from_string  ; @from ? "-infinity" : @from.to_s ; end
  def to_string    ; @to   ?  "infinity" : @to  .to_s ; end

  def include? dat
    (not @from or @from <= dat) and
      (not @to or (@exclusive ? (dat <= @to) : (dat <  @to)))
  end
  alias === include?

  private

  def mkdate d
    case d
      when nil      then nil
      when :today   then Date.today
      when Date     then d
      when Time     then Date.civil d.year, d.month, d.day
      when Fixnum   then Date.jd d
      else
        d = d.to_s
        begin
          Date.parse d
        rescue ArgumentError
          Date.strptime d, "%d.%m.%Y"
        end
    end
  end

end

