# encoding: UTF-8

#
#  collate.rb  --  Strings mit Umlauten sortieren
#

class String

  DIACRITICAL = %w(
    ÀÁÂÃÄÅÆÇÈÉÊËÌÍÎÏÐÑÒÓÔÕÖØÙÚÛÜÝÞßàáâãäåæçèéêëìíîïðñòóôõöøùúûüýþÿŒœŠšŽžſ
    AAAAAAACEEEEIIIIDNOOOOOOUUUUYtsaaaaaaaceeeeiiiidnoooooouuuuytyOoSsZzs
  ).map { |s| s.scan /./u }.transpose

  UMLAUTS = %w(ß  Æ  æ  Œ  œ ).zip \
            %w(ss ae ae OE oe)

  # The German "Telefonbuchsystem"

  UMLAUTS_TB = %w(Ä  Ö  Ü  ä  ö  ü  ß  Æ  æ  Œ  œ ).zip \
               %w(AE OE UE ae oe ue ss AE ae OE oe)

  private

  def sort_with_umlauts u
    u.each           { |u,ue| gsub! u, ue }
    DIACRITICAL.each { |d,c|  gsub! d, c  }
    gsub! /[^a-z0-9,]+/ium, " "
    gsub! %r/ *(,)/, "\\1"
    gsub! %r/  +/, " "
    strip!
    downcase!
    self
  end

  public

  # :call-seq:
  #   str.sortable!()   -> str or nil
  #
  # Reduce it to a sortable version. Examples:
  #
  #   a = "Müller, Jörg"
  #   a.sortable!
  #   a                   #=> "muller, jorg"
  #
  def sortable!
    sort_with_umlauts UMLAUTS
  end

  # :call-seq:
  #   str.sortable   -> str2
  #
  # Reduce it to a sortable version. Return a new String object without
  # touching the original one.
  #
  #   "Müller, Jörg".sortable         #=> "muller, jorg"
  #
  def sortable
    dup.sortable!
  end

  # :call-seq:
  #   str.sortable!()   -> str or nil
  #
  # Reduce it to a sortable version using the German "Telefonbuchsystem".
  # Examples:
  #
  #   a = "Jörg Müller"
  #   a.sortable_tb!
  #   a                     #=> "joerg mueller"
  #
  def sortable_tb!
    sort_with_umlauts UMLAUTS_TB
  end

  # :call-seq:
  #   str.sortable_tb   -> str2
  #
  # Reduce it to a sortable version using the German "Telefonbuchsystem".
  # Return a new String object without touching the original one.
  #
  #   "Jörg Müller".sortable         #=> "joerg mueller"
  #
  def sortable_tb
    dup.sortable_tb!
  end

end

class Array

  def sortable
    sortable_join.sortable!
  end

  def sortable_tb
    sortable_join.sortable_tb!
  end

  private

  def sortable_join
    a = dup
    a.pop until a.last or a.empty?
    a.join ", "
  end

end

