# encoding: UTF-8

#
#  deutsch/zahlwort.rb  --  Deutsche Zahlwörter generieren
#

class Zahlwort

  BEUGUNG = { nil => "", s: "s", e: "e", er: "er", es: "es" }

  attr_reader :beugung, :altmodisch

  def initialize beugung = nil, altmodisch = nil
    @beugung, @altmodisch = beugung, altmodisch
  end

  def wort i
    if i < 0 then
      "minus #{wort -i}"
    elsif i > 0 then
      r = umwandel i
      r << BEUGUNG[ @beugung] if (i % 100) == 1
      r
    else
      NULL
    end
  end

  private

  NULL = "null"

  EINER = {
     1 => "ein",
     2 => "zwei",
     3 => "drei",
     4 => "vier",
     5 => "fünf",
     6 => "sechs",
     7 => "sieben",
     8 => "acht",
     9 => "neun",
    10 => "zehn",
    11 =>"elf",
    12 =>"zwölf",
    13 =>"dreizehn",
    14 =>"vierzehn",
    15 =>"fünfzehn",
    16 =>"sechzehn",
    17 =>"siebzehn",
  }

  ZEHNER = {
    1 => "zehn",
    2 => "zwanzig",
    3 => "dreißig",
    4 => "vierzig",
    5 => "fünfzig",
    6 => "sechzig",
    7 => "siebzig",
    8 => "achtzig",
    9 => "neunzig",
  }

  HUNDERT   = "hundert"
  TAUSEND   = "tausend"
  UND       = "und"
  MILLION   = "Million"
  MILLIONEN = "Millionen"

  ZW_MIO = new :e

  def umwandel i
    r = ""
    if i == 0 then
    elsif EINER[ i] then
      r << EINER[ i]
    elsif i < 20 then
      r << EINER[ i % 10] + ZEHNER[ 1]
    elsif i < 100  then
      m = (i % 10).nonzero?
      r << EINER[ m] << UND if m
      r << ZEHNER[ i / 10]
    elsif i < 1_000 or (@altmodisch and i >= 1_100 && i < 2_000) then
      r << (umwandel i / 100) << HUNDERT << (umwandel i % 100)
    elsif i < 1_000_000 then
      r << (umwandel i / 1_000) << TAUSEND << (umwandel i % 1_000)
    else
      j = i / 1_000_000
      r << (ZW_MIO.wort j)
      n = if j == 1 then
        MILLION
      else
        MILLIONEN
      end
      r << " " << n
      u = umwandel i % 1_000_000
      r << " " << u unless u.empty?
    end
    r
  end

end

class Integer
  def zahlwort beugung = nil
    (Zahlwort.new beugung).wort self
  end
end

