# encoding: UTF-8

#
#  personen.rb  --  Eigenschaften von Personen
#

require "supplement"


class ListLine

  SEP = ";"

  class <<self

    def from_s str
      str or return
      a = split str
      new *a
    end

    private

    def split str
      a = str.split /#{self::SEP} */
      a.each { |x|
        x.gsub! /%(..)/ do ($1.to_i 0x10).chr end
      }
      a.map! { |x| x.notempty? }
      iv_propagate str, a do |t| split t end
      a
    end

    def iv_propagate str, a
      str.instance_variables.each { |iv|
        t = str.instance_variable_get iv
        if String === t then
          b = yield t
          (a.zip b).each do |(x,y)|
            x.instance_variable_set iv, y if x
          end
        end
      }
    end

  end

  def to_s
    a = to_a
    until a.empty? or a.last.notempty? do a.pop end
    a.map! { |x|
      if x then
        y = sep_mask x
        x.instance_variables.each { |iv|
          t = x.instance_variable_get iv
          if String === t then
            s = sep_mask t
            y.instance_variable_set iv, s
          end
        }
      end
      y
    }
    join_ary a, sep_sp
  end

  def inspect
    s = sep_sp
    "#<#{self.class} #{to_a.join s}>"
  end

  def join sep = nil
    join_ary to_a, sep
  end

  def == oth
    case oth
      when self.class then to_a == oth.to_a
    end
  end

  def notempty?
    to_a.any?
  end

  private

  def join_ary a, sep
    r = a.join sep
    ivs = []
    a.each { |e| ivs |= e.instance_variables }
    ivs.each { |iv|
      s = a.map { |e|
        f = e.instance_variable_get iv
        f if String === f
      }
      if s.any? then
        t = s.join sep
        r.instance_variable_set iv, t
      end
    }
    r
  end

  def sep_mask x
    x.to_s.gsub /[#{Regexp.quote self.class::SEP}%]/ do "%%%02X" % $&.ord end
  end

  def sep_sp
    "#{self.class::SEP} "
  end

end

# :class:
#   Persnam
#
# Personenname
#
class Persnam < ListLine

  class <<self
    def from_s anr, str
      str or return
      a = split str
      new anr, *a
    end
  end

  attr_reader :anr, :nachname, :vorname, :titel

  def initialize anr, *args
    @anr      = anr
    @nachname = args.shift.notempty?
    @vorname  = args.shift.notempty?
    @titel    = args.shift.notempty?
    args.empty? or
      raise ArgumentError, "Überzählige Namensbestandteile: #{args.join ','}"
  end

  def to_s
    fuer_liste
  end

  def to_a
    [ @nachname, @vorname, @titel]
  end

  def == oth
    super and @anr == oth.anr
  end

  def anr= anr
    @anr = Persnam.anr_norm anr
  end

  def anr!
    anr ||= Persnam.anr_db.anr_vorschlag @vorname
  end

  def ausfuell str
    str or return
    str.gsub /%(.)/ do
      r = case $1
        when "%" then "%"
        when "n" then @nachname
        when "v" then @vorname
        when "t" then @titel
        when "l" then fuer_liste
        when "u" then umgekehrt
        when "g" then ganz
        when "a" then anrede
      end
      if r and block_given? then yield r else r end
    end
  end
  alias expand ausfuell

  def anschreiben
    Persnam.anr_db.anschreiben anr
  end

  def grussformel &block
    ausfuell Persnam.anr_db.grussformel anr, &block
  end

  def schlussformel &block
    ausfuell Persnam.anr_db.schlussformel anr, &block
  end

  def join sep = nil
    a = to_a
    until a.last or a.empty? do a.pop end
    join_ary a, sep||", "
  end

  def fuer_liste
    r = []
    r.unshift @titel if @titel
    if @nachname =~ /\A([a-z]+ +)+/u and not $'.empty? and @vorname then
      r.unshift "#@vorname #$1"
      r.first.rstrip!
      r.unshift $'
    else
      r.unshift @vorname if @vorname or r.any?
      r.unshift @nachname
    end
    join_ary r, ", "
  end

  def umgekehrt
    r = []
    r.unshift @titel if @titel
    if @vorname then
      r.unshift @vorname
    elsif r.any?
      r.unshift ""
    end
    r.unshift @nachname if @nachname
    join_ary r, ", "
  end
  alias nach_vor umgekehrt

  def ganz
    g = [ @titel, @vorname, @nachname]
    g.compact!
    join_ary g, " "
  end

  def vorname_punkt
    @vorname.gsub /([A-ZÄÖÜ])[^ "'„“()\/:,;]+/, "\\1."
  end

  def anrede
    a = [ @titel, @nachname]
    a.compact!
    join_ary a, " "
  end

  class AnredenDB

    class <<self
      def unter? anr, muster
        anr == muster or
          (anr.starts_with? muster) && anr[ muster.length, 1] =~ /[.\/-]/i
      end
    end

    def anr_vorschlag vorn
    end
    def anr_suche anr
      case anr
        when /\A(?:[a-z_]+[.\/-])*[a-z_]+\z/i then anr
        when /^[hm]/i then "m"
        when /^[fw]/i then "w"
      end
    end
    def einzel? anr, str = nil
      anr =~ /\b[mw]\z/i or case str
        when *RE_ORG then false
        else              true
      end
    end
    RE_ORG = [
      /\bund\b/u, /&/u, /\+/u,
      /Co\b/u, /Cie\b/u, /GmbH/u,
      /kanzlei\b/ui, /büro/ui,
      /[A-Z]{2,}/u, /eV/u, /(\bund|u\.|&|\+)\s*Partner\b/u,
      %r{/}u,
      /(?:gemein|gesell|genossen|partner|wirt|land|graf)schaft/ui,
      /klinik|krankenhaus/i, /\bDres\b/u, /Vereinigung/ui
    ]

    def anschreiben anr
      anr =~ /\b[mw]\z/i and $`.notempty? and case $&.downcase
        when "m" then "Herrn"
        when "f" then "Frau"
      end
    end
    def grussformel anr
      case anr[ /\b[mw]\z/i]
        when "m" then "Sehr geehrter Herr %n"
        when "f" then "Sehr geehrte Frau %n"
        else          "Sehr geehrte Damen und Herren"
      end
    end
    def schlussformel anr
      "Mit freundlichen Grüßen"
    end
  end

  def anr_unter? muster
    Persnam.anr_unter? @anr, muster
  end

  @anr_db = AnredenDB.new

  class <<self

    attr_accessor :anr_db
    def mit_anr_db db
      old, @anr_db = @anr_db, db
      yield
    ensure
      @anr_db = old
    end

    def anr_norm anr
      @anr_db.anr_suche anr
    end

    def interpretier anr, str
      anr ||= if str.slice! /\s*\[[^\]]*\]\z/u then
        $1
      elsif str.slice! /\A\s*(Herrn?|Frau)\s+/u then
        $1
      end
      anr = anr_norm anr
      a = str_split anr, str
      iv_propagate str, a do |t| str_split anr, t end
      new *a
    end

    def anr_unter? anr, muster
      @anr_db.class.unter? anr, muster
    end

    private

    def str_split anr, str
      str = str.strip
      str.gsub! /\s+/, " "
      einz = @anr_db.einzel? anr, str
      return [ anr, str] unless einz
      case str
        when /,/              then vorn_split anr, $`,  $'
        when /\s\b[a-z]+\b.*/ then vorn_split anr, $&,  $`
        when /.*\([^\)]*\)/   then vorn_split anr, $', $&
        when /\s(\S+)$/       then vorn_split anr, $1,  $`
        else                       vorn_split anr, str, nil
      end
    end

    RE_TITEL = /^(?:
      (?:Dr|Prof|Dipl)\.(?:-Ing\.?|(\s*[a-z]+\.?)*)\s*(?:\([^)]*\)\s*)?|
      [A-Z]{2,}\.?
    )/ux

    def vorn_split a, n, vt
      n.strip! if n
      if vt =~ /,/ then
        v = $`.strip
        t = $'.strip.strip
      elsif vt then
        t = []
        while vt.slice! RE_TITEL do t.push $&.strip end
        v = vt.strip
        t = t.join " "
      end
      a ||= @anr_db.anr_vorschlag v
      [ a, n, v, t]
    end

  end

end



# :class:
#   Anschrift
#
# Postanschrift
#
class Anschrift < ListLine

  HEIM = "D"

  class <<self

    def interpretier str
      a = str.split $/
      if a.length == 1 then
        a = str.split /; */
        if a.length == 1 then
          a = str.split /, */
        end
      else
        a.reverse!
      end
      a.each { |x| x.strip! }
      if a[0] !~ /[a-z0-9]/ and (a[1] =~ /\d{4,}/ or a.length >= 3) then
        land = a.shift
      end
      if a.first =~ /\A(?:([A-Z]{1,3}) *- *)?(\d{4,})\z/ then
        plz = if not $1 or $1 == HEIM then
          $2
        else
          "#$1-#$2"
        end
        a.shift
      end
      if plz then
        ort = a.shift
        ort = [ plz, ort].join " "
        a.unshift ort
      end
      a.unshift land
      iv_propagate str, a do |t| interpretier t end
      new *a
    end

  end

  attr_reader :land, :ort, :str, :co

  def initialize *args
    l = args.first.notempty?
    unless l =~ /[0-9]/ then
      @land = l
      args.shift
    end
    @ort = args.shift.notempty?
    @str = args.shift.notempty?
    @co  = args.compact
  end

  def plz
    @ort[ /\d{4,}/]
  end

  def felder
    [ @land, @ort, @str, @co.first]
  end

  def to_a
    if @land then
      [ @land, @ort, @str, *@co]
    else
      [ @ort, @str, *@co]
    end
  end

  def konventionell
    k = to_a
    k.reverse!
    join_ary k, ", "
  end

  def zeilen
    r = to_a
    r.reverse!
    r.compact!
    r
  end

  def join sep = nil
    join_ary zeilen, sep||$/
  end

  def fuer_liste
    join_ary to_a, ", "
  end

end


# :class:
#   Bankverb
#
# Bankverbindung
#
class Bankverb < ListLine

  class <<self

    attr_writer :bank_db
    def bank_db db
      old, @bank_db = @bank_db, db
      yield
    ensure
      @bank_db = old
    end

    def interpretier str
      s = str.dup
      iban  = if s.slice! /\biban\s+(\S+)/i  then $1 end
      swift = if s.slice! /\bswift\s+(\S+)/i then $1 end
      blz   = if s.slice! /\bblz\s+(\d\d\d ?\d\d\d ?\d\d)/i then
        $1
      elsif s.slice! /[ ,;\/]*\b(\d\d\d \d\d\d \d\d)\b[ ,;\/]*/i then
        $1
      end
      if blz then
        s.slice! /(?:konto\s+)?([0-9 \/-]+\b)/i
        knr = $1
        s.slice! /\A(\s*,)*\s*/
        s.slice! /\s*(\s*,)*\z/
        bank = s
      else
        a = s.split /,/
        a.map { |x| x.strip! }
        blz = a.pick { |x| x =~ /\A\d{8}\z/ }
        knr = a.pick { |x| x =~ /\A[0-9 \/-]+/ }
        bank = a.join ", "
      end
      bank ||= @bank_db && @bank_db[ blz]
      new blz, bank, knr, iban, swift
    end

  end

  attr_reader :blz, :bank, :knr, :iban, :swift, :beschreibung
  alias kto knr

  def initialize blz, bank, knr, iban = nil, swift = nil, *rest
    @blz, @bank, @knr, @iban, @swift = blz, bank, knr, iban, swift
    @beschreibung = rest.join ", " if rest.any?
  end

  def to_a
    [ @blz, @bank, @knr, @iban, @swift]
  end

  def norm!
    @blz = @blz.gsub /\D/, ""
    @blz.insert 3, " "
    @blz.insert 7, " "
    self
  end

  def konventionell sep = nil
    r = []
    r.push "Konto #@knr"
    r.push @bank if @bank.notempty?
    r.push "BLZ #@blz"
    r.push "IBAN #@iban"   if iban
    r.push "SWIFT #@swift" if swift
    if sep then
      r.join sep
    else
      r
    end
  end

end


# :class:
#   Telkon
#
# Telekommunikationskontakte: Telefon, Fax, Netzpost
#
class Telkon < ListLine

  SEP = ","

  # :stopdoc:
  RE_TEL = /\A(?:tel\.?|(?:tele)?(?:f|ph)on|ruf|mobil):?\s*\b/i
  RE_FAX = /\A(?:tele)?fax:?\s*\b/i
  RE_WWW = /\Awww\.([a-z0-9-]+\.)+[a-z]{2,3}\z/i
  # :startdoc:

  TYPEN = [ :tel, :fax, :netzpost, :www, ]

  class <<self

    def interpretier *args
      args.each { |a|
        a or next
        (a.split /[,;\n]+/).each { |e|
          e.strip!
          case e
            when /@/               then yield :netzpost, e
            when /mailto:\b/       then yield :netzpost, $'
            when /<(.*?)>/         then yield :netzpost, $1
            when %r<http://\b>     then yield :www,      $'
            when RE_WWW            then yield :www,      $&
            when RE_TEL            then yield :tel,      $'
            when RE_FAX            then yield :fax,      $'
            when %r<[0-9()/ -]+\z> then yield :tel,      e
            else                        yield nil,       e
          end
        }
      }
      nil
    end

    def als_hash *args
      t = {}
      Telkon.interpretier *args do |typ,wert|
        typ ||= yield wert if block_given?
        typ or raise "Kann `#{wert}' keinem Telekommunikationstyp zuordnen."
        t[ typ] ||= Telkon.new
        t[ typ].hinzu wert
      end
      t
    end

  end

  attr_reader :nummern

  def initialize *nummern
    nummern.flatten!
    @nummern = nummern.map { |n| n.to_s.notempty? }
    @nummern.compact!
  end

  def to_a
    @nummern
  end

  def liste
    to_a.join ", "
  end

  def == oth
    case oth
      when self.class then @nummern.sort == oth.nummern.sort
    end
  end

  def & oth
    @nummern & oth.nummern
  end

  def | oth
    @nummern | oth.nummern
  end

  def < oth
    @nummern.each { |n|
      return false unless oth.nummern.include? n
    }
    true
  end

  def > oth
    oth < self
  end

  def notempty?
    @nummern.notempty? and self
  end

  def hat? num
    @nummern[ typ].include? num
  end
  alias include? hat?

  def [] i
    @nummern[ i]
  end

  def hinzu num
    @nummern.push num if num.notempty?
  end
  alias << hinzu

end

